<%------------------------------------------------------------------------------
\section{Driver}
%------------------------------------------------------------------------------

\paragraph{Usage}

\begin{verbatim}
alex Lexer.x
frown --lexer --expected --code=stackless --optimize Parser.lg

ghci -Wall -package rh Driver.lhs
ghc --make -package rh -package util Driver.lhs
ghc --make Driver.lhs
\end{verbatim}

%if False
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Module header}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> module Main
> where
> import            System.IO
> import            System.Exit (exitWith, ExitCode(..))
> import            System.Environment (getArgs, getEnv)
> import            Data.List
> import            Data.Char (isSpace)
> import            Data.IORef
> import            Control.Monad (when, foldM)
> import qualified  System.Console.Readline as Readline
> import qualified  System.Console.SimpleLineEditor as SLE
> import            Control.Concurrent
> import            Control.Exception
> import            Prelude hiding (catch)
>
> import            Config
> import            Options
> import            Message hiding (concat, break, intersperse)
> import            Parser (parsePhrase, parseExpr, parseDecls)
> import            Core hiding (($)) --(expr2core, decls2core)
> import            Evaluate hiding (split) -- (execute)
> import            TypeCheck
> import            AbstractSyntax (pType, ppSpec, Spec(..), unknown, Type(..), SigExpr (Sig))
> import            Interrupt

%endif

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Driver}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> bpl :: Options -> IO ()
> bpl options
>   =  catch (
>      do
>      -- readline setup
>         when (interactive options && readline options) $
>           do  Readline.initialize
>      -- The following is for Windows and hopefully does no harm on Unix;
>      -- for some reason, readline on Windows seems to see the AltGr key
>      -- as Meta, causing problems entering several characters on international
>      -- keyboards. We just add some additional key combinations to fix this.
>               Readline.parseAndBind "\"\\M-\\\\\": \"\\\\\""
>               Readline.parseAndBind "\"\\M-[\": \"[\""
>               Readline.parseAndBind "\"\\M-]\": \"]\""
>               Readline.parseAndBind "\"\\M-{\": \"{\""
>               Readline.parseAndBind "\"\\M-}\": \"}\""
>               Readline.parseAndBind "\"\\M-~\": \"~\""
>               Readline.parseAndBind "\"\\M-|\": \"|\""
>               Readline.setCompletionEntryFunction (Just (const (return [])))
>               Readline.setCompletionAppendCharacter Nothing
>      -- automatically activate interactive mode if no files are given
>         options <- if (null (files options) && not (interactive options))
>                    then do hPutStrLn stderr "Warning: no input files. Assuming --interactive."
>                            return (options { interactive = True })
>                    else return options
>      -- print welcome message
>         when (interactive options && not (quiet options))
>           (putStr (welcomeTxt ++ "Type :? for help\n"))
>      -- read history
>         when (not (null (history options)) && readline options) (
>           do  contents <- getFile options (history options)
>               sequence_ [Readline.addHistory s | s <- lines contents])
>      -- prevent the Windows 2000 bug
>         options <- return $ options { files = filter (not . ("\\" `isPrefixOf`)) (files options) }
>
>      -- read BPL_LIBDIR from environment
>         libdir <- catch  (getEnv "BPL_LIBDIR" >>= return . filter (/='"'))
>                          (\_ -> hPutStrLn stderr "Warning: BPL_LIBDIR not set, assuming current directory ..." >> return ".")
>         options <- return $ options { searchpath = [ libdir, libdir ++ "/Library", "." ] }
>      -- add System.bpl
>         options <- return $ options { files = if bare options then files options else "System.bpl" : (files options) }
>
>         let cmds   =  [ Compile (CompileFile f) | f <- files options ]
>      -- "interactive" is temporarily disabled during loading the pre-specified files
>         let inter  =  interactive options
>         let state  =  (options { interactive = False },initenv,(0,[]),inittpenv,[],0)
>         state <- foldM (\s c -> handleCommand s c >>= \ (Just s) -> return s) state cmds
>         let state' =  case state of
>                         (options,env,mem,te,sub,cnt) -> (options { interactive = inter },env,mem,te,sub,cnt)
>      -- install signal handler
>         t <- myThreadId
>         when (interactive options) (installHandler (handler t) >> return ())
>      -- read-eval-print loop
>         if interactive options then
>           let readevalprint :: DriverState -> IO ()
>               readevalprint state@(options, env, mem, tp, sub, cnt) =
>                 catch (unblock
>                   (do when (readline options) (Readline.setCompletionEntryFunction (Just (makeCompletion env tp)))
>                       inp <- if readline options then  -- read
>                                 Readline.readline "BPL > "
>                              else
>                                 do
>                                   SLE.initialise
>                                   x <- SLE.getLineEdited "BPL > "
>                                   SLE.restore
>                                   return x
>                       case inp of
>                         Nothing -> return () -- EOF encountered
>                         Just s ->
>                           do  when (not (null s) && readline options) (Readline.addHistory s)
>                               c   <-  interpretCommand s
>                               st  <-  handleCommand state c
>                               case st of
>                                 Nothing  ->  return () -- quit encountered
>                                 Just state'
>                                          ->  readevalprint state'))
>                   -- We use a recursive exception handler here: if multiple interrupts
>                   -- are issued during evaluation, the threadDelay will cause those
>                   -- to be raised immediately (if we don't do this, the next readline
>                   -- call might block exception handling, and we'll only get interrupted
>                   -- during evaluation of the *next* expression again.
>                   (let k b e =
>                             catch
>                               (case e of
>                                 Deadlock -> 
>                                   do
>                                     when b $ hPutStrLn stderr "Interrupted."
>                                     unblock $ threadDelay 1000 >> readevalprint state
>                                 _ ->
>                                   case userErrors e of
>                                     Just _ -> unblock $ readevalprint state
>                                     _      -> print e >> unblock (readevalprint state))
>                               (k False)
>                    in k True)
>           in  do
>                 -- when (not (null te)) $ putStrLn (render (Page pageWidth) (stack (map ppSpec (reverse te)))) -- print type info
>                 readevalprint state'
>          else
>           return ()
>      ) (\ e -> case userErrors e of
>                  Just _ -> hPutStrLn stderr "compilation aborted"
>                  _      -> hPutStrLn stderr ("bplc: " ++ show e)) >>
>      when (readline options) (Readline.resetTerminal Nothing)
>   where
>   checkExpr :: TypeEnv -> TSubst -> Int -> String -> IO (Maybe Type)
>   checkExpr te0 sub cnt raw
>     =  do  when (verbose options) (eject (Title (text ("parsing " ++ "<interactive>"))))
>            str         <-  pass 1 (input raw)
>            ast         <-  pass 2 (parseExpr str)
>            t           <-  fmap (fmap (\ (x,_,_) -> x)) (runTC options sub cnt (typecheck te0 ast))
>            return t
>
>   compileFile :: Options -> TypeEnv -> TSubst -> Int -> String -> IO (Either (Expr -> Expr,[Spec],TSubst,Int) (Expr,Maybe (Type,TSubst,Int)))
>   compileFile options te0 sub cnt fileName
>     =  do  raw <- getFile options fileName
>            -- shebang convention
>            let raw' = if take 2 raw == "#!" then dropWhile (/='\n') raw else raw
>            compilePhrase te0 sub cnt raw' fileName 
>
>   compilePhrase :: TypeEnv -> TSubst -> Int -> String -> String -> IO (Either (Expr -> Expr,[Spec],TSubst,Int) (Expr,Maybe (Type,TSubst,Int)))
>   compilePhrase te0 sub cnt raw fileName
>     =  do  when (verbose options) (eject (Title (text ("compiling " ++ fileName))))
>            str         <-  pass 1 (input raw)
>            ast         <-  pass 2 (parsePhrase str)
>            case ast of
>              Left   ds  ->
>                do  (te1,sub',cnt')  <-  fmap (maybe ([],sub,cnt) id) (runTC options sub cnt (typecheckDecls te0 ds))
>                    core  <-  pass 4 (decls2core ds)
>                    return (Left (core, te1, sub', cnt'))
>              Right  e   ->
>                do  t     <-  runTC options sub cnt (typecheck te0 e)
>                    core  <-  pass 4 (expr2core e)
>                    return (Right (core, t))
>
>   handleCommand :: DriverState -> Command -> IO (Maybe DriverState)
>   handleCommand state@(options, env, mem, te, sub, cnt) cmd
>     =  case cmd of
>          Quit   ->  return Nothing
>          Noop   ->  return (Just state)
>          Set option
>            | not (option `isPrefixOf` "") && option `isPrefixOf` "trace"
>                 ->  do  let n = not (trace options)
>                         putStrLn $ "trace now " ++ if n then "on" else "off"
>                         return (Just (options { trace = n }, env, mem, te, sub, cnt))
>            | not (option `isPrefixOf` "s") && option `isPrefixOf` "signature"
>                 ->  do  let n = not (signature options)
>                         putStrLn $ "signature now " ++ if n then "on" else "off"
>                         return (Just (options { signature = n }, env, mem, te, sub, cnt))
>            | not (option `isPrefixOf` "s") && option `isPrefixOf` "stricttc"
>                 ->  do  let n = not (stricttc options)
>                         putStrLn $ "stricttc now " ++ if n then "on" else "off"
>                         return (Just (options { stricttc = n }, env, mem, te, sub, cnt))
>            | otherwise
>                 ->  putStrLn ("unknown or ambiguous option; currently supported: trace, signature") >> return (Just state)
>          Help   ->  putStr (helpTxt commands) >> return (Just state)
>          TypeOf x ->
>                     do  t <- checkExpr te sub cnt x
>                         maybe (return ()) (\u -> putStrLn (render (Page pageWidth) (pType u))) t
>                         return (Just state)
>          Browse ->  do  if (typechecker options) then
>                           putStr (render  (Page pageWidth)
>                                           (stack  (map ppSpec (reverse (nubBy (\x y -> specName x == specName y) te)) 
>                                                   ++ [empty])))
>                          else
>                           putStr (render  (Page pageWidth)
>                                           (stack  (map text (reverse (nub (map fst env)))
>                                                   ++ [empty])))
>                         return (Just state)
>          Reload ->  if null (files options) then
>                       return (Just state)
>                     else
>                       handleCommand state (Compile (CompileFile (last (files options))))
>          Compile f ->
>                        do  p <-  catch (  fmap Just $
>                                           case f of
>                                             CompileInteractive s  ->  compilePhrase te sub cnt s "<interactive>"
>                                             CompileFile f         ->  compileFile options te sub cnt f )
>                                        (\ _ -> return Nothing)
>                            let options'  =  case f of
>                                               CompileFile f  ->  options { files = files options ++ [f] }
>                                               _              ->  options
>                            -- putStrLn "compilation complete"
>                            case p of
>                               Nothing -> return (Just (options', env, mem, te, sub, cnt))
>                               Just (Left (ds,te',sub',cnt')) ->
>                                     -- bindings
>                                     do cfg2 <- execute options (ds Struct) env mem
>                                        -- putStrLn "execution complete (decls)"
>                                        let te'' = te' ++ te
>                                        case cfg2 of Stop (VStruct env') mem' ->
>                                                          do
>                                                            when (not (null te') && interactive options && signature options) $
>                                                              putStr (render (Page pageWidth) (stack (map ppSpec (reverse te') ++ [empty]))) -- print type info
>                                                            return (Just (options', env', mem', te'', sub', cnt'))
>                                                     _ -> do
>                                                            putStrLn (render (Page pageWidth) (ppState cfg2))  -- print error
>                                                            return (Just (options', env, mem, te, sub, cnt))
>                               Just (Right (e,t)) ->
>                                     -- expression
>                                     do cfg2 <- execute options e env mem                  -- eval
>                                        -- putStrLn "execution complete (expr)"
>                                        putStrLn (render (Page pageWidth) (ppState cfg2))  -- print
>                                        case cfg2 of Stop v mem' -> do
>                                                                      let  env'           =  ("it",v):env
>                                                                           (b,sub',cnt')  =  maybe ([],sub,cnt) (\(u,s,c) -> ([SpVal (unknown,"it") u],s,c)) t
>                                                                           te'            =  b ++ te
>                                                                      when (interactive options && signature options) (mapM_ (\u -> putStrLn (render (Page pageWidth) (ppSpec u))) b)
>                                                                      return (Just (options', env', mem', te', sub', cnt'))
>                                                     _ -> return (Just (options', env, mem, te, sub, cnt))
>
>   initenv    =  stdenv
>   inittpenv  =  if not (typechecker options) then [] else stdtypeenv
>
>   dump = dumpphases options
>   pass n (Fail msg) 
>     =  do  eject msg
>            fail "compilation aborted"
>   pass n (Succeed a msg) 
>     =  do  when (n `elem` dump) (eject msg)
>            return a
>   pageWidth = pagewidth options
>   eject a = putStrLn (render (Page pageWidth) (pretty a))
>
>   -- to prevent problems on Windows with locked files, we read all files strictly;
>   -- otherwise, we might keep a lock on a partially read file, for instance if a parse
>   -- error occurs early in a file
>   getFile options x
>     =  let  tryFile []      =  do  hPutStrLn stderr ("Error: File `" ++ x ++ "' not found. Ignoring ...")
>                                    return ""
>             tryFile (p:ps)  =  catch (strictReadFile p) (const (tryFile ps))
>        in   tryFile (x : map (\p -> p ++ "/" ++ x) (searchpath options))
>
>   -- Reads a file completely into memory.
>   strictReadFile :: FilePath -> IO String
>   strictReadFile f  =   do  f <- readFile f
>                             f `stringSeq` return f
>
>   -- Completely evaluates a string.
>   stringSeq :: String -> b -> b
>   stringSeq []      c  =  c
>   stringSeq (x:xs)  c  =  stringSeq xs c


>
> data Command = TypeOf String
>              | Compile CompileForm
>              | Reload
>              | Browse
>              | Trace
>              | Set String
>              | Quit
>              | Help
>              | Noop

> data CompileForm = CompileInteractive  String
>                  | CompileFile         String

> data InteractiveCommand = Cmd [String] String (String -> Command) String

> commands
>   =  [ Cmd [":type"]        "<expr>"  TypeOf         "print type of expression",
>        Cmd [":browse"]      ""        (const Browse) "browse names in scope",
>        Cmd [":debug"]       ""        (const (Set "trace")) "toggle tracer",
>        Cmd [":load"]        "<file>"  (Compile . CompileFile)
>                                                      "load program from file",
>        Cmd [":reload"]      ""        (const Reload) "reload the file last loaded",
>        Cmd [":set"]         "<option>" Set           "toggle option",
>        Cmd [":quit"]        ""        (const Quit)   "exit BPL interpreter",
>        Cmd [":help",":?"]   ""        (const Help)   "display this list of commands" ]

> interpretCommand :: String -> IO Command
> interpretCommand x
>   =  if isPrefixOf ":" x then
>        do  let  (cmd,t')  =  break isSpace x
>                 t         =  dropWhile isSpace t'
>            -- find matching commands
>            let  matching  =  filter (\ (Cmd cs _ _ _) -> any (isPrefixOf cmd) cs) commands
>            case matching of
>              []  ->  do  putStrLn ("Unknown command `" ++ cmd ++ "'. Type :? for help.")
>                          return Noop
>              (Cmd _ _ f _ : _)
>                  ->  do  return (f t)
>      else
>        return (Compile (CompileInteractive x))

> type DriverState = (Options, Env, Mem, TypeEnv, TSubst, Int)

> welcomeTxt :: String
> welcomeTxt
>   =  "bplc (version " ++ versionTxt ++ "), Copyright 2006 Ralf Hinze and Andres Loeh\n"

> helpTxt :: [InteractiveCommand] -> String
> helpTxt cs
>   =  "List of commands:  Any command may be abbreviated to :c where\n\
>      \c is the first character in the full name.\n\n\
>      \<expr>              evaluate expression\n\
>      \<decl>              evaluate declaration\n"
>      ++
>      unlines (map (\ (Cmd cs a _ d) -> let  ct = concat (intersperse ", " (map (++ if null a then "" else " " ++ a) cs))
>                                        in   ct ++ replicate ((20 - length ct) `max` 2) ' ' ++ d) cs)

> main
>   =  do  argv <- getArgs
>          case getOptions argv of
>            (options, fnames, [])
>              |  copying options                         ->  putStrLn copyingTxt
>              |  warranty options                        ->  putStrLn warrantyTxt
>              |  help options                            ->  putStr (usageInfo header optionDescrs)
>              |  version options                         ->  putStrLn licenseTxt
>              |  otherwise                               ->  bpl (options { files = fnames })
>            (_, _, errors)                               ->  do  putStrLn (concat errors ++ usageInfo header optionDescrs)
>                                                                 exitWith (ExitFailure 1)
>          exitWith ExitSuccess
>   where header = "Usage: bplc [option ...] file ...\n"

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pass 1}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> input raw  =  Succeed raw msg
>   where msg = [ Subtitle (text "reader"),
>                 Debug (text "source"),
>                 Debug (text ("\n" ++ raw)) ]

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{readline completion}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data CEnv = CEnv { unCEnv :: [(String, Maybe CEnv)] }

> makeCompletion :: Env -> TypeEnv -> String -> IO [String]
> makeCompletion e tp x = 
>     let
>       tre  =  CEnv . map (\ (x,y) -> (x, case y of VStruct s -> Just (tre s); _ -> Nothing))
>       trp  =  CEnv . map (\ s -> case s of SpModule (_,n) (Sig t) -> (n, Just (trp t)); _ -> (snd (specName s), Nothing))
>       e'   =  tre e
>       tp'  =  trp tp
>       xs   =  CEnv (keywords ++ unCEnv (if null tp then e' else tp'))
>       qs = split '.' x
>       (m,n) =  case qs of
>                  []  ->  ([],"")
>                  _   ->  (init qs,last qs)
>     in
>       -- hPutStrLn stderr ("finding completion for: " ++ x) >>
>       return (concatMap (process m n) (unCEnv (look m xs)))
>   where  -- unfortunately neither lexer nor parser seem to export a list of keywords
>          keywords =
>            [ (n,Nothing) | n <-
>              [  "and", "array", "case", "catch", "class", "continue",
>                 "data", "do", "else", "end", "exception", "for", "forall",
>                 "fun", "function", "functor", "if", "in", "inherit", "interface",
>                 "let", "letcc", "local", "method", "module", "new", "object",
>                 "of", "open", "rec", "ref", "sig", "struct", "then", "throw",
>                 "try", "type", "val", "value", "while" ] ]
>
>          look :: [String] -> CEnv -> CEnv
>          look [] e               =  e
>          look (q:qs) (CEnv e)    =  case lookup q e of
>                                       Just (Just e')  ->  look qs e'
>                                       _               ->  CEnv []
>
>          process m n (x,Just _)  =  process' m n " " x ++ process' m n "." x
>          process m n (x,v)       =  process' m n " " x
>          
>          process' m n c x 
>            | isPrefixOf n x      =  [concat (intersperse "." (m ++ [x])) ++ c]
>            | otherwise           =  []

Variant of |words| that accepts a custom separator.

> split :: Char -> String -> [String]
> split c s =  case dropWhile (==c) s of
>                ""  | null s     ->  []
>                    | otherwise  ->  [[]]
>                s'  ->  w : split c s''
>                           where  (w, s'') = break (==c) s'

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Signal handler}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> handler t = throwTo t Deadlock -- abusing deadlock for user interrupt

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{License and warranty}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> licenseTxt :: String
> licenseTxt
>   =  welcomeTxt ++ "\n\
>      \bplc comes with ABSOLUTELY NO WARRANTY; for details type `bplc --warranty'.\n\
>      \This is free software, and you are welcome to redistribute it\n\
>      \under certain conditions; type `bplc --copying' for details."

> copyingTxt :: String
> copyingTxt
>   =  "\t\t    GNU GENERAL PUBLIC LICENSE\n\
>      \\t\t       Version 2, June 1991\n\
>      \\n\
>      \ Copyright (C) 1989, 1991 Free Software Foundation, Inc.,\n\
>      \ 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n\
>      \ Everyone is permitted to copy and distribute verbatim copies\n\
>      \ of this license document, but changing it is not allowed.\n\
>      \\n\
>      \\t\t\t    Preamble\n\
>      \\n\
>      \  The licenses for most software are designed to take away your\n\
>      \freedom to share and change it.  By contrast, the GNU General Public\n\
>      \License is intended to guarantee your freedom to share and change free\n\
>      \software--to make sure the software is free for all its users.  This\n\
>      \General Public License applies to most of the Free Software\n\
>      \Foundation's software and to any other program whose authors commit to\n\
>      \using it.  (Some other Free Software Foundation software is covered by\n\
>      \the GNU Library General Public License instead.)  You can apply it to\n\
>      \your programs, too.\n\
>      \\n\
>      \  When we speak of free software, we are referring to freedom, not\n\
>      \price.  Our General Public Licenses are designed to make sure that you\n\
>      \have the freedom to distribute copies of free software (and charge for\n\
>      \this service if you wish), that you receive source code or can get it\n\
>      \if you want it, that you can change the software or use pieces of it\n\
>      \in new free programs; and that you know you can do these things.\n\
>      \\n\
>      \  To protect your rights, we need to make restrictions that forbid\n\
>      \anyone to deny you these rights or to ask you to surrender the rights.\n\
>      \These restrictions translate to certain responsibilities for you if you\n\
>      \distribute copies of the software, or if you modify it.\n\
>      \\n\
>      \  For example, if you distribute copies of such a program, whether\n\
>      \gratis or for a fee, you must give the recipients all the rights that\n\
>      \you have.  You must make sure that they, too, receive or can get the\n\
>      \source code.  And you must show them these terms so they know their\n\
>      \rights.\n\
>      \\n\
>      \  We protect your rights with two steps: (1) copyright the software, and\n\
>      \(2) offer you this license which gives you legal permission to copy,\n\
>      \distribute and/or modify the software.\n\
>      \\n\
>      \  Also, for each author's protection and ours, we want to make certain\n\
>      \that everyone understands that there is no warranty for this free\n\
>      \software.  If the software is modified by someone else and passed on, we\n\
>      \want its recipients to know that what they have is not the original, so\n\
>      \that any problems introduced by others will not reflect on the original\n\
>      \authors' reputations.\n\
>      \\n\
>      \  Finally, any free program is threatened constantly by software\n\
>      \patents.  We wish to avoid the danger that redistributors of a free\n\
>      \program will individually obtain patent licenses, in effect making the\n\
>      \program proprietary.  To prevent this, we have made it clear that any\n\
>      \patent must be licensed for everyone's free use or not licensed at all.\n\
>      \\n\
>      \  The precise terms and conditions for copying, distribution and\n\
>      \modification follow.\n\
>      \\f\n\
>      \\t\t    GNU GENERAL PUBLIC LICENSE\n\
>      \   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n\
>      \\n\
>      \  0. This License applies to any program or other work which contains\n\
>      \a notice placed by the copyright holder saying it may be distributed\n\
>      \under the terms of this General Public License.  The \"Program\", below,\n\
>      \refers to any such program or work, and a \"work based on the Program\"\n\
>      \means either the Program or any derivative work under copyright law:\n\
>      \that is to say, a work containing the Program or a portion of it,\n\
>      \either verbatim or with modifications and/or translated into another\n\
>      \language.  (Hereinafter, translation is included without limitation in\n\
>      \the term \"modification\".)  Each licensee is addressed as \"you\".\n\
>      \\n\
>      \Activities other than copying, distribution and modification are not\n\
>      \covered by this License; they are outside its scope.  The act of\n\
>      \running the Program is not restricted, and the output from the Program\n\
>      \is covered only if its contents constitute a work based on the\n\
>      \Program (independent of having been made by running the Program).\n\
>      \Whether that is true depends on what the Program does.\n\
>      \\n\
>      \  1. You may copy and distribute verbatim copies of the Program's\n\
>      \source code as you receive it, in any medium, provided that you\n\
>      \conspicuously and appropriately publish on each copy an appropriate\n\
>      \copyright notice and disclaimer of warranty; keep intact all the\n\
>      \notices that refer to this License and to the absence of any warranty;\n\
>      \and give any other recipients of the Program a copy of this License\n\
>      \along with the Program.\n\
>      \\n\
>      \You may charge a fee for the physical act of transferring a copy, and\n\
>      \you may at your option offer warranty protection in exchange for a fee.\n\
>      \\n\
>      \  2. You may modify your copy or copies of the Program or any portion\n\
>      \of it, thus forming a work based on the Program, and copy and\n\
>      \distribute such modifications or work under the terms of Section 1\n\
>      \above, provided that you also meet all of these conditions:\n\
>      \\n\
>      \    a) You must cause the modified files to carry prominent notices\n\
>      \    stating that you changed the files and the date of any change.\n\
>      \\n\
>      \    b) You must cause any work that you distribute or publish, that in\n\
>      \    whole or in part contains or is derived from the Program or any\n\
>      \    part thereof, to be licensed as a whole at no charge to all third\n\
>      \    parties under the terms of this License.\n\
>      \\n\
>      \    c) If the modified program normally reads commands interactively\n\
>      \    when run, you must cause it, when started running for such\n\
>      \    interactive use in the most ordinary way, to print or display an\n\
>      \    announcement including an appropriate copyright notice and a\n\
>      \    notice that there is no warranty (or else, saying that you provide\n\
>      \    a warranty) and that users may redistribute the program under\n\
>      \    these conditions, and telling the user how to view a copy of this\n\
>      \    License.  (Exception: if the Program itself is interactive but\n\
>      \    does not normally print such an announcement, your work based on\n\
>      \    the Program is not required to print an announcement.)\n\
>      \\f\n\
>      \These requirements apply to the modified work as a whole.  If\n\
>      \identifiable sections of that work are not derived from the Program,\n\
>      \and can be reasonably considered independent and separate works in\n\
>      \themselves, then this License, and its terms, do not apply to those\n\
>      \sections when you distribute them as separate works.  But when you\n\
>      \distribute the same sections as part of a whole which is a work based\n\
>      \on the Program, the distribution of the whole must be on the terms of\n\
>      \this License, whose permissions for other licensees extend to the\n\
>      \entire whole, and thus to each and every part regardless of who wrote it.\n\
>      \\n\
>      \Thus, it is not the intent of this section to claim rights or contest\n\
>      \your rights to work written entirely by you; rather, the intent is to\n\
>      \exercise the right to control the distribution of derivative or\n\
>      \collective works based on the Program.\n\
>      \\n\
>      \In addition, mere aggregation of another work not based on the Program\n\
>      \with the Program (or with a work based on the Program) on a volume of\n\
>      \a storage or distribution medium does not bring the other work under\n\
>      \the scope of this License.\n\
>      \\n\
>      \  3. You may copy and distribute the Program (or a work based on it,\n\
>      \under Section 2) in object code or executable form under the terms of\n\
>      \Sections 1 and 2 above provided that you also do one of the following:\n\
>      \\n\
>      \    a) Accompany it with the complete corresponding machine-readable\n\
>      \    source code, which must be distributed under the terms of Sections\n\
>      \    1 and 2 above on a medium customarily used for software interchange; or,\n\
>      \\n\
>      \    b) Accompany it with a written offer, valid for at least three\n\
>      \    years, to give any third party, for a charge no more than your\n\
>      \    cost of physically performing source distribution, a complete\n\
>      \    machine-readable copy of the corresponding source code, to be\n\
>      \    distributed under the terms of Sections 1 and 2 above on a medium\n\
>      \    customarily used for software interchange; or,\n\
>      \\n\
>      \    c) Accompany it with the information you received as to the offer\n\
>      \    to distribute corresponding source code.  (This alternative is\n\
>      \    allowed only for noncommercial distribution and only if you\n\
>      \    received the program in object code or executable form with such\n\
>      \    an offer, in accord with Subsection b above.)\n\
>      \\n\
>      \The source code for a work means the preferred form of the work for\n\
>      \making modifications to it.  For an executable work, complete source\n\
>      \code means all the source code for all modules it contains, plus any\n\
>      \associated interface definition files, plus the scripts used to\n\
>      \control compilation and installation of the executable.  However, as a\n\
>      \special exception, the source code distributed need not include\n\
>      \anything that is normally distributed (in either source or binary\n\
>      \form) with the major components (compiler, kernel, and so on) of the\n\
>      \operating system on which the executable runs, unless that component\n\
>      \itself accompanies the executable.\n\
>      \\n\
>      \If distribution of executable or object code is made by offering\n\
>      \access to copy from a designated place, then offering equivalent\n\
>      \access to copy the source code from the same place counts as\n\
>      \distribution of the source code, even though third parties are not\n\
>      \compelled to copy the source along with the object code.\n\
>      \\f\n\
>      \  4. You may not copy, modify, sublicense, or distribute the Program\n\
>      \except as expressly provided under this License.  Any attempt\n\
>      \otherwise to copy, modify, sublicense or distribute the Program is\n\
>      \void, and will automatically terminate your rights under this License.\n\
>      \However, parties who have received copies, or rights, from you under\n\
>      \this License will not have their licenses terminated so long as such\n\
>      \parties remain in full compliance.\n\
>      \\n\
>      \  5. You are not required to accept this License, since you have not\n\
>      \signed it.  However, nothing else grants you permission to modify or\n\
>      \distribute the Program or its derivative works.  These actions are\n\
>      \prohibited by law if you do not accept this License.  Therefore, by\n\
>      \modifying or distributing the Program (or any work based on the\n\
>      \Program), you indicate your acceptance of this License to do so, and\n\
>      \all its terms and conditions for copying, distributing or modifying\n\
>      \the Program or works based on it.\n\
>      \\n\
>      \  6. Each time you redistribute the Program (or any work based on the\n\
>      \Program), the recipient automatically receives a license from the\n\
>      \original licensor to copy, distribute or modify the Program subject to\n\
>      \these terms and conditions.  You may not impose any further\n\
>      \restrictions on the recipients' exercise of the rights granted herein.\n\
>      \You are not responsible for enforcing compliance by third parties to\n\
>      \this License.\n\
>      \\n\
>      \  7. If, as a consequence of a court judgment or allegation of patent\n\
>      \infringement or for any other reason (not limited to patent issues),\n\
>      \conditions are imposed on you (whether by court order, agreement or\n\
>      \otherwise) that contradict the conditions of this License, they do not\n\
>      \excuse you from the conditions of this License.  If you cannot\n\
>      \distribute so as to satisfy simultaneously your obligations under this\n\
>      \License and any other pertinent obligations, then as a consequence you\n\
>      \may not distribute the Program at all.  For example, if a patent\n\
>      \license would not permit royalty-free redistribution of the Program by\n\
>      \all those who receive copies directly or indirectly through you, then\n\
>      \the only way you could satisfy both it and this License would be to\n\
>      \refrain entirely from distribution of the Program.\n\
>      \\n\
>      \If any portion of this section is held invalid or unenforceable under\n\
>      \any particular circumstance, the balance of the section is intended to\n\
>      \apply and the section as a whole is intended to apply in other\n\
>      \circumstances.\n\
>      \\n\
>      \It is not the purpose of this section to induce you to infringe any\n\
>      \patents or other property right claims or to contest validity of any\n\
>      \such claims; this section has the sole purpose of protecting the\n\
>      \integrity of the free software distribution system, which is\n\
>      \implemented by public license practices.  Many people have made\n\
>      \generous contributions to the wide range of software distributed\n\
>      \through that system in reliance on consistent application of that\n\
>      \system; it is up to the author/donor to decide if he or she is willing\n\
>      \to distribute software through any other system and a licensee cannot\n\
>      \impose that choice.\n\
>      \\n\
>      \This section is intended to make thoroughly clear what is believed to\n\
>      \be a consequence of the rest of this License.\n\
>      \\f\n\
>      \  8. If the distribution and/or use of the Program is restricted in\n\
>      \certain countries either by patents or by copyrighted interfaces, the\n\
>      \original copyright holder who places the Program under this License\n\
>      \may add an explicit geographical distribution limitation excluding\n\
>      \those countries, so that distribution is permitted only in or among\n\
>      \countries not thus excluded.  In such case, this License incorporates\n\
>      \the limitation as if written in the body of this License.\n\
>      \\n\
>      \  9. The Free Software Foundation may publish revised and/or new versions\n\
>      \of the General Public License from time to time.  Such new versions will\n\
>      \be similar in spirit to the present version, but may differ in detail to\n\
>      \address new problems or concerns.\n\
>      \\n\
>      \Each version is given a distinguishing version number.  If the Program\n\
>      \specifies a version number of this License which applies to it and \"any\n\
>      \later version\", you have the option of following the terms and conditions\n\
>      \either of that version or of any later version published by the Free\n\
>      \Software Foundation.  If the Program does not specify a version number of\n\
>      \this License, you may choose any version ever published by the Free Software\n\
>      \Foundation.\n\
>      \\n\
>      \  10. If you wish to incorporate parts of the Program into other free\n\
>      \programs whose distribution conditions are different, write to the author\n\
>      \to ask for permission.  For software which is copyrighted by the Free\n\
>      \Software Foundation, write to the Free Software Foundation; we sometimes\n\
>      \make exceptions for this.  Our decision will be guided by the two goals\n\
>      \of preserving the free status of all derivatives of our free software and\n\
>      \of promoting the sharing and reuse of software generally.\n\
>      \\n"
>      ++ warrantyTxt ++
>      "\n\n\
>      \\t\t     END OF TERMS AND CONDITIONS\n\
>      \\f\n\
>      \\t    How to Apply These Terms to Your New Programs\n\
>      \\n\
>      \  If you develop a new program, and you want it to be of the greatest\n\
>      \possible use to the public, the best way to achieve this is to make it\n\
>      \free software which everyone can redistribute and change under these terms.\n\
>      \\n\
>      \  To do so, attach the following notices to the program.  It is safest\n\
>      \to attach them to the start of each source file to most effectively\n\
>      \convey the exclusion of warranty; and each file should have at least\n\
>      \the \"copyright\" line and a pointer to where the full notice is found.\n\
>      \\n\
>      \    <one line to give the program's name and a brief idea of what it does.>\n\
>      \    Copyright (C) 19yy  <name of author>\n\
>      \\n\
>      \    This program is free software; you can redistribute it and/or modify\n\
>      \    it under the terms of the GNU General Public License as published by\n\
>      \    the Free Software Foundation; either version 2 of the License, or\n\
>      \    (at your option) any later version.\n\
>      \\n\
>      \    This program is distributed in the hope that it will be useful,\n\
>      \    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
>      \    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
>      \    GNU General Public License for more details.\n\
>      \\n\
>      \    You should have received a copy of the GNU General Public License\n\
>      \    along with this program; see the file COPYING.  If not, write to\n\
>      \    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,\n\
>      \    Boston, MA 02111-1307, USA.\n\
>      \\n\
>      \Also add information on how to contact you by electronic and paper mail.\n\
>      \\n\
>      \If the program is interactive, make it output a short notice like this\n\
>      \when it starts in an interactive mode:\n\
>      \\n\
>      \    Gnomovision version 69, Copyright (C) 19yy name of author\n\
>      \    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.\n\
>      \    This is free software, and you are welcome to redistribute it\n\
>      \    under certain conditions; type `show c' for details.\n\
>      \\n\
>      \The hypothetical commands `show w' and `show c' should show the appropriate\n\
>      \parts of the General Public License.  Of course, the commands you use may\n\
>      \be called something other than `show w' and `show c'; they could even be\n\
>      \mouse-clicks or menu items--whatever suits your program.\n\
>      \\n\
>      \You should also get your employer (if you work as a programmer) or your\n\
>      \school, if any, to sign a \"copyright disclaimer\" for the program, if\n\
>      \necessary.  Here is a sample; alter the names:\n\
>      \\n\
>      \  Yoyodyne, Inc., hereby disclaims all copyright interest in the program\n\
>      \  `Gnomovision' (which makes passes at compilers) written by James Hacker.\n\
>      \\n\
>      \  <signature of Ty Coon>, 1 April 1989\n\
>      \  Ty Coon, President of Vice\n\
>      \\n\
>      \This General Public License does not permit incorporating your program into\n\
>      \proprietary programs.  If your program is a subroutine library, you may\n\
>      \consider it more useful to permit linking proprietary applications with the\n\
>      \library.  If this is what you want to do, use the GNU Library General\n\
>      \Public License instead of this License."

> warrantyTxt :: String
> warrantyTxt
>   =  "\t\t\t    NO WARRANTY\n\
>      \\n\
>      \  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n\
>      \FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n\
>      \OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n\
>      \PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n\
>      \OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n\
>      \MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n\
>      \TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n\
>      \PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n\
>      \REPAIR OR CORRECTION.\n\
>      \\n\
>      \  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n\
>      \WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n\
>      \REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n\
>      \INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n\
>      \OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n\
>      \TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n\
>      \YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n\
>      \PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n\
>      \POSSIBILITY OF SUCH DAMAGES."